from __future__ import division, print_function
import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg
np.set_printoptions(precision=12)
from control import *

#Pick example from thermal, turbine1 and turbine2
Example = 'thermal'

#dlqr idea based on here: http://www.mwm.im/lqr-controllers-with-python/
def dlqr(A,B,Q,R):
    #ref Bertsekas, p.151
    #X = np.matrix(scipy.linalg.solve_discrete_are(A, B, Q, R))
    if abs(R[0]) < 1.0e-12:
        R[0] = 1.0e-12
    print(R);
    X,L,E = dare(A,B,Q,R)
    #compute the LQR gain
    K = np.matrix(scipy.linalg.inv(B.T*X*B+R)*(B.T*X*A))
    return K, X

# One state turbine example
if Example == 'turbine1':
    T = 0.01
    A = np.matrix([[.9999]])
    B = np.matrix([[T]])
    C = np.matrix([1])
    D = np.matrix([0])

# Two state turbine example
if Example == 'turbine2':
    T = 0.01
    A = np.matrix([[1.0,T],[0, 1]])
    B = np.matrix([[0],[T]])
    C = np.matrix([1,0])
    D = np.matrix([0])

# Thermal example (normalized)
if Example == 'thermal':
    T = 1
    g = 0.1
    A = np.matrix([[1-2*g*T,g*T,0,0],[g*T,1-2*g*T,g*T,0],[0,g*T,1-2*g*T,g*T],[0,0,g*T,1-g*T]])
    B = np.matrix([[g*T],[0],[0],[0]])
    C = np.matrix([0,0,0,1])
    D = np.matrix([0])

print("A matrix:")
print(A)
print("B matrix:")
print(B)


# Problem number of states
N = A.shape[0]

#cost matrices for LQR.  REMEMBER PYTHON INDEXES FROM ZERO
# The weight for the first state is Q[0,0].
Q = np.eye(N)
Q[0,0] = 2
R = 1*np.eye(1)
print("Q:");
print(Q);
k_lqr,x = dlqr(A,B,Q,R)

#Natural frequency picking
nat_freqs = np.zeros(N)
k_acker = acker(A,B,nat_freqs)

for k in [k_acker, k_lqr]:
    #Print k
    print("K values")
    print(k)

    #calculate K_u:
    KuInv = C*np.linalg.inv(np.eye(N)-(A-B*k))*B
    Ku = np.linalg.inv(KuInv)
    print(['Ku=' + str(Ku)])
    
    #Form closed loop matrix 
    Ac = A-B*k #equivalent to M in notes
    Bc = B*Ku
    Cc = C
    Dc = D
    overall=ss(Ac,Bc,Cc,Dc,T)

    #Print closed-loop eigenvalues and estimate simulation interval
    print("Eigenvalues and abs")
    ec, evecc = scipy.linalg.eig(Ac)
    max_n = 100
    for p in ec:
        abs_p = np.abs(p)
        print([p, abs_p])
        max_n = max(max_n, 2*np.log(0.1)/np.log(abs_p+1.0e-12))

    #Simulate the system
    print('Simulation Time Interval = '+str(max_n*T))
    t = np.arange(0,max_n*T,T)
    if Example == 'thermal':
        sig = 20*np.ones(np.size(t))
    else:
        sig = np.ones(np.size(t))
    y,t1,x=lsim(overall,sig,t)
            
    #Plot the results
    for i in range(N):
        plt.subplot(N+1, 1, i+1)
        plt.plot(x[i]);
        plt.ylabel('State '+str(i+1))

    command = np.array(Ku*sig[1] - k*x)
    plt.subplot(N+1, 1, N+1)
    plt.plot(command[0])
    plt.ylabel('u')
    plt.xlabel('time steps')

    if np.all(k == k_acker):
        plt.suptitle('States and Input with gains from placing natural frequencies');
    else:
        plt.suptitle('State and Input with gains from LQR');
    plt.show()

